@@ -1,4 +1,4 @@
-/* $Id: gucharmap-window.c,v 1.18 2003/06/04 23:12:56 nlevitt Exp $ */
+/* $Id: gucharmap-window.c,v 1.28 2003/11/18 16:30:12 nlevitt Exp $ */
 /*
  * Copyright (c) 2003  Noah Levitt <nlevitt аt columbia.edu>
  *
@@ -94,101 +94,6 @@ jump_clipboard (GtkWidget *widget, GucharmapWindow *guw)
 }
 
 
-static void
-jump_code_point_response (GtkDialog *dialog, 
-                          gint response, 
-                          EntryDialog *entry_dialog)
-{
-  if (response == GTK_RESPONSE_OK)
-    {
-      const gchar *text;
-      gchar *message;
-      gchar *endptr;
-      glong l;
-
-      text = gtk_entry_get_text (entry_dialog->entry);
-
-      l = strtol (text, &endptr, 16);
-
-      if (*text != '\0' && *endptr == '\0' && l >= 0 && l <= UNICHAR_MAX)
-        gucharmap_charmap_go_to_character (entry_dialog->guw->charmap, 
-                                           (gunichar) l);
-      else
-        {
-          message = g_strdup_printf (_("Not a valid code point to jump to."
-                                       " Must be a hexadecimal number between"
-                                       " 0 and %4.4X."), UNICHAR_MAX);
-          information_dialog (entry_dialog->guw, GTK_WINDOW (dialog), message);
-          g_free (message);
-          return;
-        }
-    }
-
-  g_free (entry_dialog);
-  gtk_widget_destroy (GTK_WIDGET (dialog));
-}
-
-
-static void
-jump_code_point (GtkWidget *widget, GucharmapWindow *guw)
-{
-  GtkWidget *dialog;
-  GtkWidget *entry;
-  GtkWidget *label;
-  GtkWidget *crapbox;
-  GtkWidget *spacer;
-  EntryDialog *entry_dialog;
-
-  dialog = gtk_dialog_new_with_buttons (_("Jump to Unicode Code Point"),
-                                        GTK_WINDOW (guw),
-                                        GTK_DIALOG_DESTROY_WITH_PARENT
-                                        | GTK_DIALOG_NO_SEPARATOR, 
-                                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, 
-                                        GTK_STOCK_OK, GTK_RESPONSE_OK, 
-                                        NULL);
-
-  crapbox = gtk_vbox_new (FALSE, 0);
-  gtk_widget_show (crapbox);
-  gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), crapbox);
-
-  gtk_window_set_icon (GTK_WINDOW (dialog), guw->icon);
-  gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);
-  gtk_container_set_border_width (GTK_CONTAINER (crapbox), 6);
-  gtk_box_set_spacing (GTK_BOX (crapbox), 6);
-
-  label = gtk_label_new_with_mnemonic (_("_Enter hexadecimal Unicode"
-                                         " code point:"));
-  gtk_widget_show (label);
-  gtk_box_pack_start (GTK_BOX (crapbox), label, FALSE, FALSE, 0);
-
-  entry = gtk_entry_new ();
-  gtk_widget_show (entry);
-  gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
-  gtk_entry_set_max_length (GTK_ENTRY (entry), 8); 
-  gtk_entry_set_width_chars (GTK_ENTRY (entry), 8);
-  gtk_box_pack_start (GTK_BOX (crapbox), entry, FALSE, FALSE, 0);
-
-  spacer = gtk_alignment_new (0, 0, 0, 0);
-  gtk_widget_show (spacer);
-  gtk_widget_set_size_request (spacer, -1, 6);
-  gtk_box_pack_start (GTK_BOX (crapbox), spacer, FALSE, FALSE, 0);
-
-  gtk_label_set_mnemonic_widget (GTK_LABEL (label), entry);
-
-  entry_dialog = g_new (EntryDialog, 1); 
-  entry_dialog->guw = guw;
-  entry_dialog->dialog = dialog;
-  entry_dialog->entry = GTK_ENTRY (entry);
-
-  g_signal_connect (GTK_DIALOG (dialog), "response", 
-                    G_CALLBACK (jump_code_point_response), entry_dialog);
-
-  gtk_widget_show_all (dialog);
-
-  gtk_widget_grab_focus (entry);
-}
-
-
 static void
 status_message (GtkWidget *widget, const gchar *message, GucharmapWindow *guw)
 {
@@ -206,7 +111,80 @@ do_search (GucharmapWindow *guw,
            const gchar *search_text, 
            gint direction)
 {
+  const gchar *no_leading_space, *nptr;
+  char *endptr;
   g_assert (direction == -1 || direction == 1);
+  gunichar wc;
+
+  if (search_text[0] == '\0')
+    {
+      information_dialog (guw, alert_parent, _("Nothing to search for."));
+      return FALSE;
+    }
+
+  /* skip spaces */
+  for (no_leading_space = search_text;  
+       g_unichar_isspace (g_utf8_get_char (no_leading_space));
+       no_leading_space = g_utf8_next_char (no_leading_space));
+
+  /* check for explicit decimal codepoint */
+  nptr = no_leading_space;
+  if (strncasecmp (no_leading_space, "&#", 2) == 0)
+    nptr = no_leading_space + 2;
+  else if (*no_leading_space == '#')
+    nptr = no_leading_space + 1;
+
+  if (nptr != no_leading_space)
+    {
+      wc = strtoul (nptr, &endptr, 10);
+      if (endptr != nptr && wc <= UNICHAR_MAX)
+        {
+          gucharmap_charmap_go_to_character (guw->charmap, wc);
+          return TRUE;
+        }
+    }
+
+  /* check for explicit hex code point */
+  nptr = no_leading_space;
+  if (strncasecmp (no_leading_space, "&#x", 3) == 0)
+    nptr = no_leading_space + 3;
+  else if (strncasecmp (no_leading_space, "U+", 2) == 0
+           || strncasecmp (no_leading_space, "0x", 2) == 0)
+    nptr = no_leading_space + 2;
+
+  if (nptr != no_leading_space)
+    {
+      wc = strtoul (nptr, &endptr, 16);
+      if (endptr != nptr && wc <= UNICHAR_MAX)
+        {
+          gucharmap_charmap_go_to_character (guw->charmap, wc);
+          return TRUE;
+        }
+    }
+
+  /* if there is only one character and it isspace() jump to it */
+  if (g_utf8_get_char (search_text) <= UNICHAR_MAX
+      && g_utf8_strlen (search_text, -1) == 1)
+    {
+      gucharmap_charmap_go_to_character (guw->charmap, g_utf8_get_char (search_text));
+      return TRUE;
+    }
+
+  /* “Unicode character names contain only uppercase Latin letters A through
+   * Z, digits, space, and hyphen-minus.” If first character is not one of
+   * those, jump to it. Also, if there is only one character, jump to it.
+   * */
+  if (g_utf8_get_char (no_leading_space) <= UNICHAR_MAX
+      && (g_utf8_strlen (no_leading_space, -1) == 1
+          || ! ((*no_leading_space >= 'A' && *no_leading_space <= 'Z') 
+                || (*no_leading_space >= 'a' && *no_leading_space <= 'z') 
+                || (*no_leading_space >= '0' && *no_leading_space <= '9') 
+                || (*no_leading_space == '-')
+                || (*no_leading_space == '\0'))))
+    {
+      gucharmap_charmap_go_to_character (guw->charmap, g_utf8_get_char (no_leading_space));
+      return TRUE;
+    }
 
   switch (gucharmap_charmap_search (guw->charmap, search_text, direction))
     {
@@ -215,7 +193,33 @@ do_search (GucharmapWindow *guw,
         return TRUE;
 
       case GUCHARMAP_NOT_FOUND:
-        information_dialog (guw, alert_parent, _("Not found."));
+        /* try searching without leading spaces */
+        switch (gucharmap_charmap_search (guw->charmap, no_leading_space, direction))
+          {
+            case GUCHARMAP_FOUND:
+            case GUCHARMAP_WRAPPED:
+              return TRUE;
+
+            case GUCHARMAP_NOT_FOUND:
+              /* NOW try interpreting it from the start as a hex codepoint */
+              wc = strtoul (no_leading_space, &endptr, 16);
+              if (endptr != nptr && wc <= UNICHAR_MAX)
+                {
+                  gucharmap_charmap_go_to_character (guw->charmap, wc);
+                  return TRUE;
+                }
+
+              information_dialog (guw, alert_parent, _("Not found."));
+              return FALSE;
+
+            case GUCHARMAP_NOTHING_TO_SEARCH_FOR:
+              information_dialog (guw, alert_parent, _("Nothing to search for."));
+              return FALSE;
+
+            default:
+              g_warning ("gucharmap_charmap_search returned an unexpected result; this should never happen");
+              return FALSE;
+          }
         return FALSE;
 
       case GUCHARMAP_NOTHING_TO_SEARCH_FOR:
@@ -364,25 +368,40 @@ snap_cols_pow2 (GtkCheckMenuItem *mi, GucharmapWindow *guw)
 static void
 help_about (GtkWidget *widget, GucharmapWindow *guw)
 {
-  GtkWidget *about;
-  const gchar *authors[] = { "Noah Levitt <nlevitt аt columbia.edu>", 
-                             "Daniel Elstner <daniel.elstner аt gmx.net>", 
-                             "Padraig O'Briain <Padraig.Obriain аt sun.com>",
-                             NULL };
-  const gchar *translator_credits;
+  static GtkWidget *about = NULL;
+  if (about == NULL) 
+    {
+      const gchar *authors[] = 
+        { 
+          "Noah Levitt <nlevitt аt columbia.edu>", 
+          "Daniel Elstner <daniel.elstner аt gmx.net>", 
+          "Padraig O'Briain <Padraig.Obriain аt sun.com>",
+          NULL 
+        };
+      const gchar *documenters[] =
+	{
+	  "Chee Bin HOH <cbhoh аt gnome.org>",
+          NULL
+	};	  
+      const gchar *translator_credits;
 
-  translator_credits = _("translator_credits");
-  if (strcmp (translator_credits, "translator_credits") == 0)
-    translator_credits = NULL;
+      translator_credits = _("translator_credits");
+      if (strcmp (translator_credits, "translator_credits") == 0)
+        translator_credits = NULL;
 
-  about = gnome_about_new (
-          "gucharmap", VERSION, 
-          "Copyright © 2003 Noah Levitt <nlevitt аt columbia.edu>",
-          _("Unicode Character Map"), authors, NULL, 
-          translator_credits, guw->icon);
+      about = gnome_about_new (
+              "gucharmap", VERSION, 
+              "Copyright © 2003 Noah Levitt <nlevitt аt columbia.edu>",
+              _("Unicode Character Map"), authors, documenters,
+              translator_credits, guw->icon);
 
-  gtk_window_set_icon (GTK_WINDOW (about), guw->icon);
-  gtk_widget_show (about);
+      /* set the widget pointer to NULL when the widget is destroyed */
+      g_signal_connect (G_OBJECT (about), "destroy",
+                        G_CALLBACK (gtk_widget_destroyed), &about);
+      gtk_window_set_icon (GTK_WINDOW (about), guw->icon);
+    }
+
+  gtk_window_present (GTK_WINDOW (about));
 }
 
 
@@ -391,6 +410,7 @@ make_gnome_help_menu (GucharmapWindow *guw)
 {
   GnomeUIInfo help_menu[] =
   {
+    GNOMEUIINFO_HELP (PACKAGE),
     GNOMEUIINFO_MENU_ABOUT_ITEM (help_about, guw),
     GNOMEUIINFO_END
   };
@@ -398,7 +418,7 @@ make_gnome_help_menu (GucharmapWindow *guw)
 
   menu = gtk_menu_new ();
 
-  gnome_app_fill_menu (GTK_MENU_SHELL (menu), help_menu, NULL, TRUE, 0);
+  gnome_app_fill_menu (GTK_MENU_SHELL (menu), help_menu, guw->accel_group, TRUE, 0);
 
   return menu;
 }
@@ -415,10 +435,8 @@ history_back (GtkWidget *widget,
   guw->history = guw->history->prev;
 
   gtk_widget_set_sensitive (guw->back_menu_item, guw->history->prev != NULL);
-  gtk_widget_set_sensitive (guw->back_button, guw->history->prev != NULL);
 
   gtk_widget_set_sensitive (guw->forward_menu_item, TRUE);
-  gtk_widget_set_sensitive (guw->forward_button, TRUE);
 
   gucharmap_table_set_active_character (guw->charmap->chartable,
                                         (gunichar) guw->history->data);
@@ -436,10 +454,8 @@ history_forward (GtkWidget *widget,
 
   gtk_widget_set_sensitive (guw->forward_menu_item, 
                             guw->history->next != NULL);
-  gtk_widget_set_sensitive (guw->forward_button, guw->history->next != NULL);
 
   gtk_widget_set_sensitive (guw->back_menu_item, TRUE);
-  gtk_widget_set_sensitive (guw->back_button, TRUE);
 
   gucharmap_table_set_active_character (guw->charmap->chartable,
                                         (gunichar) guw->history->data);
@@ -610,13 +626,6 @@ make_menu (GucharmapWindow *guw)
   /* separator */
   gtk_menu_shell_append (GTK_MENU_SHELL (search_menu), gtk_menu_item_new ());
 
-  menu_item = gtk_menu_item_new_with_mnemonic (_("Code _Point..."));
-  gtk_widget_add_accelerator (menu_item, "activate", guw->accel_group,
-                              GDK_j, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
-  g_signal_connect (G_OBJECT (menu_item), "activate",
-                    G_CALLBACK (jump_code_point), guw);
-  gtk_menu_shell_append (GTK_MENU_SHELL (search_menu), menu_item);
-
   menu_item = gtk_menu_item_new_with_mnemonic (_("Character in _Clipboard"));
   gtk_widget_add_accelerator (menu_item, "activate", guw->accel_group,
                               GDK_o, GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);
@@ -849,44 +858,7 @@ link_clicked (GucharmapCharmap *charmap,
   guw->history = guw->history->next;
 
   gtk_widget_set_sensitive (guw->back_menu_item, TRUE);
-  gtk_widget_set_sensitive (guw->back_button, TRUE);
   gtk_widget_set_sensitive (guw->forward_menu_item, FALSE);
-  gtk_widget_set_sensitive (guw->forward_button, FALSE);
-}
-
-static GtkWidget *
-button_new_with_stock_image (gchar *message,
-                             gchar *stock_id)
-{
-  GtkWidget *button;
-  GtkWidget *hbox;
-  GtkWidget *image;
-  GtkWidget *label;
-  GtkWidget *align;
-
-  button = gtk_button_new ();
-  gtk_widget_show (button);
-
-  label = gtk_label_new (message);
-  gtk_widget_show (label);
-
-  image = gtk_image_new_from_stock (stock_id, GTK_ICON_SIZE_BUTTON);
-  gtk_widget_show (image);
-
-  hbox = gtk_hbox_new (FALSE, 3);
-  gtk_widget_show (hbox);
-
-  gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);
-  gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
-
-  align = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
-  gtk_widget_show (align);
-
-  gtk_container_add (GTK_CONTAINER (align), hbox);
-
-  gtk_container_add (GTK_CONTAINER (button), align);
-
-  return button;
 }
 
 
@@ -895,8 +867,6 @@ pack_stuff_in_window (GucharmapWindow *guw)
 {
   GtkWidget *big_vbox;
   GtkWidget *hbox;
-  GtkWidget *button_box;
-  GtkWidget *separator;
 
   guw->charmap = GUCHARMAP_CHARMAP (gucharmap_charmap_new ());
 
@@ -917,31 +887,6 @@ pack_stuff_in_window (GucharmapWindow *guw)
   gtk_box_pack_start (GTK_BOX (hbox), guw->fontsel, FALSE, FALSE, 0);
   gtk_widget_show (GTK_WIDGET (guw->charmap));
 
-  separator = gtk_alignment_new (0, 0, 0, 0);
-  gtk_widget_show (separator);
-  gtk_widget_set_size_request (separator, 12, -1);
-  gtk_box_pack_start (GTK_BOX (hbox), separator, FALSE, FALSE, 0);
-
-  button_box = gtk_hbutton_box_new ();
-  gtk_widget_show (button_box);
-  gtk_box_pack_start (GTK_BOX (hbox), button_box, FALSE, FALSE, 0);
-
-  guw->back_button = button_new_with_stock_image (_("Back"), 
-                                                  GTK_STOCK_GO_BACK);
-  gtk_widget_show (guw->back_button);
-  gtk_widget_set_sensitive (guw->back_button, FALSE);
-  gtk_container_add (GTK_CONTAINER (button_box), guw->back_button);
-  g_signal_connect (G_OBJECT (guw->back_button), "clicked",
-                    G_CALLBACK (history_back), guw);
-
-  guw->forward_button = button_new_with_stock_image (_("Forward"), 
-                                                  GTK_STOCK_GO_FORWARD);
-  gtk_widget_show (guw->forward_button);
-  gtk_widget_set_sensitive (guw->forward_button, FALSE);
-  gtk_container_add (GTK_CONTAINER (button_box), guw->forward_button);
-  g_signal_connect (G_OBJECT (guw->forward_button), "clicked",
-                    G_CALLBACK (history_forward), guw);
-
   guw->text_to_copy_container = make_text_to_copy (guw);
   gtk_container_set_border_width (GTK_CONTAINER (guw->text_to_copy_container), 
                                   6);
@@ -994,10 +939,29 @@ show_all (GtkWidget *widget)
 }
 
 
+static void
+window_finalize (GObject *object)
+{
+  GucharmapWindow *guw = GUCHARMAP_WINDOW (object);
+
+  if (guw->last_search)
+    g_free (guw->last_search);
+
+  if (guw->history)
+    {
+      while (guw->history->prev)
+        guw->history = guw->history->prev;
+
+      g_list_free (guw->history);
+    }
+}
+
+
 static void
 gucharmap_window_class_init (GucharmapWindowClass *clazz)
 {
   GTK_WIDGET_CLASS (clazz)->show_all = show_all;
+  G_OBJECT_CLASS (clazz)->finalize = window_finalize;
 }
 
 
